"""The monoSpace function receives a string and returns it with all
consecutive white spaces replaced by a single space. Only characters that
are recognized as digits, letters or punctuation are included. Include in
the string the following tags to explicitly set new lines or indentations:
  '<br>' for new lines
  '<tab>' for one tab containing the number of spaces defined at the
  keyword argument 'tab', by default 2."""
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from string import digits, punctuation, ascii_letters


def _getPrintables() -> str:
  """Return a list of printable characters."""
  return '%s %s %s' % (digits, punctuation, ascii_letters)


def _removeNonPrintableCharacters(word: str) -> str:
  """Remove all non-printable characters from the string."""
  printables = _getPrintables()
  return ''.join([c if c in printables else ' ' for c in word])


def monoSpace(words: str, **kwargs) -> str:
  """The monoSpace function receives a string and returns it with all
  consecutive white spaces replaced by a single space. Only characters that
  are recognized as digits, letters or punctuation are included. Include in
  the string the following tags to explicitly set new lines or indentations:
    '<br>' for new lines
    '<n: int>' for indentations of 'n' spaces.
    '<tab>' for one tab containing the number of spaces defined at the
    keyword argument 'tab', by default 2."""
  tabSymbol = kwargs.get('tab', '<tab>')
  newLineSymbol = kwargs.get('newLine', '<br>')
  words = _removeNonPrintableCharacters(words)
  lines = words.split(newLineSymbol)
  lines = [' '.join(line.split()) for line in lines]
  lines = [line.replace(tabSymbol, ' ' * 2) for line in lines]
  return '\n'.join(lines)



"""WorkTest provides a general subclass of TestCase"""
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from unittest import TestCase


class WorkTest(TestCase):
  """WorkTest provides a general subclass of TestCase for testing."""

  def __getattribute__(self, key: str) -> object:
    """Careful now!"""
    val = object.__getattribute__(self, key)
    if 'test' in key and False:
      if callable(val):
        clsName = self.__class__.__name__
        print('%s.%s' % (clsName, val.__name__))
    return val

  @classmethod
  def setUpClass(cls, ) -> None:
    """Prints the name of the class"""
    # print('\nBeginning test of class: %s!' % cls.__name__)



"""EmptyField provides a flexible implementation of the descriptor
protocol allowing owning classes to decorate methods as accessor methods. """
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from typing import Callable

from worktoy.text import monoSpace, typeMsg


class EmptyField:
  """EmptyField provides a flexible implementation of the descriptor
  protocol allowing owning classes to decorate methods as accessor
  methods. """
  __field_name__ = None
  __field_owner__ = None
  __field_type__ = None
  __getter_name__ = None
  __setter_name__ = None
  __deleter_name__ = None

  def _getFieldName(self) -> str:
    """Returns the name of the field the descriptor is assigned to. """
    if self.__field_name__ is None:
      e = """The descriptor has not been assigned to a field. """
      raise AttributeError(monoSpace(e))
    if isinstance(self.__field_name__, str):
      return self.__field_name__
    e = typeMsg('__field_name__', self.__field_name__, str)
    raise TypeError(monoSpace(e))

  def _getFieldOwner(self) -> type:
    """Returns the type of the class the descriptor is assigned to. """
    if self.__field_owner__ is None:
      e = """The descriptor has not been assigned to a field. """
      raise AttributeError(monoSpace(e))
    if isinstance(self.__field_owner__, type):
      return self.__field_owner__
    e = typeMsg('__field_owner__', self.__field_owner__, type)
    raise TypeError(monoSpace(e))

  def _getOwnerName(self) -> str:
    """Returns the name of the class the descriptor is assigned to. """
    return self._getFieldOwner().__name__

  def _getFieldType(self) -> type:
    """Getter-function for the field type."""
    if self.__field_type__ is None:
      return object
    if isinstance(self.__field_type__, type):
      return self.__field_type__
    e = typeMsg('__field_type__', self.__field_type__, type)
    raise TypeError(e)

  def __set_name__(self, owner: type, name: str) -> None:
    """Set the name of the field and the owner of the field."""
    self.__field_name__ = name
    self.__field_owner__ = owner

  def __get__(self, instance: object, owner: type) -> object:
    """Get the value of the field."""
    if instance is None:
      return self
    if not issubclass(owner, self._getFieldOwner()):
      e = """The instance does not belong to the owner class: '%s'."""
      raise RuntimeError(monoSpace(e % self._getOwnerName()))
    fieldName = self._getFieldName()
    ownerName = self._getOwnerName()
    if self.__getter_name__ is None:
      e = """The field instance at name: '%s' does not have a getter!"""
      raise AttributeError(monoSpace(e % fieldName))
    getter = getattr(owner, self.__getter_name__, None)
    if getter is None:
      e = """The owner class: '%s' does not implement a getter function
      for field: '%s'."""
      raise AttributeError(monoSpace(e % (ownerName, fieldName)))
    if callable(getter):
      return getter(instance)
    e = typeMsg('getter', getter, Callable)
    raise TypeError(e)

  def __set__(self, instance: object, value: object) -> None:
    """Set the value of the field."""
    fieldName = self._getFieldName()
    ownerName = self._getOwnerName()
    if not isinstance(instance, self.__field_owner__):
      e = """The instance does not belong to the owner class: '%s'."""
      raise RuntimeError(monoSpace(e % ownerName))
    if self.__setter_name__ is None:
      e = """The field instance at name: '%s' does not have a setter!"""
      raise AttributeError(monoSpace(e % fieldName))
    setter = getattr(self.__field_owner__, self.__setter_name__, None)
    if setter is None:
      e = """The owner class: '%s' does not implement a setter function
      for field: '%s'."""
      raise AttributeError(monoSpace(e % (ownerName, fieldName)))
    if callable(setter):
      return setter(instance, value)
    e = typeMsg('setter', setter, Callable)
    raise TypeError(e)

  def __delete__(self, instance: object) -> None:
    """Delete the value of the field."""
    fieldName = self._getFieldName()
    ownerName = self._getOwnerName()
    if not isinstance(instance, self.__field_owner__):
      e = """The instance does not belong to the owner class: '%s'."""
      raise RuntimeError(monoSpace(e % ownerName))
    if self.__deleter_name__ is None:
      e = """The field instance at name: '%s' does not have a deleter!"""
      raise AttributeError(monoSpace(e % fieldName))
    deleter = getattr(self.__field_owner__, self.__deleter_name__, None)
    if deleter is None:
      e = """The owner class: '%s' does not implement a deleter function
      for field: '%s'."""
      raise AttributeError(monoSpace(e % (ownerName, fieldName)))
    if callable(deleter):
      return deleter(instance)
    e = typeMsg('deleter', deleter, Callable)
    raise TypeError(e)

  def __set_getter__(self, callMeMaybe: Callable) -> Callable:
    """Set the getter function of the field."""
    self.__getter_name__ = callMeMaybe.__name__
    return callMeMaybe

  def __set_setter__(self, callMeMaybe: Callable) -> Callable:
    """Set the setter function of the field."""
    self.__setter_name__ = callMeMaybe.__name__
    return callMeMaybe

  def __set_deleter__(self, callMeMaybe: Callable) -> Callable:
    """Set the deleter function of the field."""
    self.__deleter_name__ = callMeMaybe.__name__
    return callMeMaybe

  def GET(self, callMeMaybe: Callable) -> Callable:
    """Decorator for setting the getter function of the field."""
    return self.__set_getter__(callMeMaybe)

  def SET(self, callMeMaybe: Callable) -> Callable:
    """Decorator for setting the setter function of the field."""
    return self.__set_setter__(callMeMaybe)

  def DELETE(self, callMeMaybe: Callable) -> Callable:
    """Decorator for setting the deleter function of the field."""
    return self.__set_deleter__(callMeMaybe)

  def DEL(self, callMeMaybe: Callable) -> Callable:
    """Decorator for setting the deleter function of the field."""
    return self.__set_deleter__(callMeMaybe)



"""This file provides the BeginFields and EndFields tags zerotons. These
zerotons may be placed in the top of the class body of a new EZData class.
Between these two tags, all names are expected to be automatic fields.
These must implement the descriptor protocol. """
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from worktoy.meta import ZerotonMetaclass


class BeginFields(metaclass=ZerotonMetaclass):
  """Marks the beginning of the fields section in the body of an EZData
  class."""


class EndFields(metaclass=ZerotonMetaclass):
  """Marks the end of the fields section in the body of an EZData class."""



"""DataMetaclass provides a custom metaclass for dataclasses. Classes
derived from DataMetaclass should use instances of AttriBox in their class
body to denote data fields. These fields are populated automatically by
the __init__ methods. Derived classes are free to defined methods,
but must not implement the __init__ method as it is reserved for
populating the data fields."""
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from worktoy.meta import AbstractMetaclass
from worktoy.ezdata import DataNamespace


class DataMetaclass(AbstractMetaclass):
  """DataMetaclass provides a custom metaclass for dataclasses. Classes
  derived from DataMetaclass should use instances of AttriBox in their class
  body to denote data fields. These fields are populated automatically by
  the __init__ methods. Derived classes are free to defined methods,
  but must not implement the __init__ method as it is reserved for
  populating the data fields."""

  @classmethod
  def __prepare__(mcls, name: str, _, **kwargs) -> DataNamespace:
    return DataNamespace(mcls, name, (), **kwargs)

  def __str__(cls) -> str:
    """String representation"""
    return cls.__name__


class EZData(metaclass=DataMetaclass):
  """EZData provides a base class for dataclasses. Data fields must be
  instances of AttriBox and the __init__ method is created automatically.
  Subclasses are not allowed to implement the __init__ method. """



"""The 'runTests' function runs tests found in the tests directory."""
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

import os
import unittest


def runTests() -> int:
  """Runs the tests"""
  results = []
  loader = unittest.TestLoader()
  res = None
  for item in os.listdir('tests'):
    if not item.startswith('test'):
      continue
    testPath = os.path.join('tests', item)
    suite = loader.discover(start_dir=testPath, pattern='test*.py')
    runner = unittest.TextTestRunner(verbosity=0)
    res = runner.run(suite)
    if res.wasSuccessful():
      results.append('Tests passed in: %s' % testPath)
    else:
      results.append('Tests failed in: %s' % testPath)
  for result in results:
    print(result)
  if res is None:
    return -1
  return 0



"""The ZerotonMetaclass provides a custom metaclass for creation of Token
objects. They are like singletons, but since they never instantiate,
they are called zerotons. """
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from typing import Never

from worktoy.meta import AbstractMetaclass


class ZerotonMetaclass(AbstractMetaclass):
  """ZerotonMetaclass provides a custom metaclass for creation of Token
  objects. They are like singletons, but since they never instantiate,
  they are called zerotons. """

  def __new__(mcls, name: str, _, space: dict[str, object]) -> type:
    """The __new__ method is invoked to create the class."""
    cls = super().__new__(mcls, name, (), space)
    return cls

  def __call__(cls, *__, **_) -> Never:
    """The __call__ method is invoked when the class is called."""
    e = """%s is a zeroton and cannot be instantiated!""" % cls.__name__
    raise TypeError(e)

  def __eq__(cls, other: object) -> bool:
    """The __eq__ method is invoked when the class is compared
    with another object."""
    if isinstance(other, str):
      return False if cls.__name__ != other else True
    return False if hash(cls) - hash(other) else True

  def __hash__(cls, ) -> int:
    """Obtaining the hash value of the class name"""
    return str.__hash__(cls.__name__)



"""The 'yolo' function receives any number of callables and runs them."""
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

import sys
import time
from typing import Callable

from icecream import ic

from worktoy.math import log


def yolo(*args: Callable) -> None:
  """The 'yolo' function receives any number of callables and runs them."""
  tic = time.perf_counter_ns()
  print('Running python script located at: \n%s' % sys.argv[0])
  print('Started at: %s' % time.ctime())
  print(77 * '-')
  retCode = 0
  for callMeMaybe in args:
    print('\nRunning: %s\n' % callMeMaybe.__name__)
    try:
      retCode = callMeMaybe()
    except BaseException as exception:
      print('Exception: %s' % exception)
      retCode = -1
  retCode = 0 if retCode is None else retCode
  print(77 * '-')
  print('Return Code: %s' % retCode)
  toc = int(time.perf_counter_ns() - tic)
  n = 0
  while toc > 1e03:
    toc *= 1e-03
    n += 1
  name = ['nano', 'micro', 'milli', ''][int(n)]
  print('Runtime: %d %s-seconds' % (int(toc), name))



"""The 'typeMsg' function generates a structured error message when an
object with a given name does not belong to the given type.

Example:

  def square(number: int) -> int:
    #  The function expects an integer as argument
    if not isinstance(number, int):
      e = typeMsg('number', number, int)
      raise TypeError(e)
    return number ** 2

  square(69.420)



"""
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from worktoy.text import monoSpace


def typeMsg(name: str, obj: object, type_: type) -> str:
  """The 'typeMsg' function generates a structured error message when an
  object with a given name does not belong to the given type."""
  expectedType = type_.__name__
  actualType = type(obj).__name__
  e = """Expected object '%s' to be of type '%s', but found '%s' of type 
  '%s'!"""
  return monoSpace(e % (name, expectedType, str(obj), actualType))



"""The 'worktoy.meta' package provides base classes for working with
metaclasses.

Consider the following class definition:
class SomeClass:
  pass

The above code is equivalent to the following:
class SomeClass(object, metaclass=type):
  pass

The 'metaclass' defaults to 'type' and every object in python is an
instance of 'object', so the above code is equivalent to the first. A
custom metaclass subclasses 'type' allowing for customizing everything about
what it means to be a class in python. This concept is the most powerful tool
in python. This is because any potentially more powerful tool can be
achieved by using a custom metaclass.

Below is an example of a custom metaclass which explicitly does what type
does during the class creation process. This illustrates the power of the
custom metaclass, because each step can be entirely customized.


class MetaClass(type):
  #  MetaClass inherits from 'type' making it a usable as a metaclass.

  @classmethod
  def __prepare__(mcls,
                  name: str,
                  bases: tuple[type, ...],
                  **kwargs) -> dict:
    #  The __prepare__ method is a class method meaning that the first
    #  argument it receives, is the metaclass itself. The classmethod
    # decorator is required by the __prepare__ method. This method creates
    #  the namespace object used during the class creation process. The
    #  default namespace object is an empty instance of 'dict'.
    return dict()

  def __new__(mcls,
              name: str,
              bases: tuple[type, ...],
              namespace: dict,
              **kwargs) -> type:
    #  The __new__ method is a class method meaning that the first
    #  argument it receives, is the metaclass itself. This method creates
    #  the class object. The default class object is an instance of 'type'.
    #  Please note, that while this method is indeed a class method, it must
    #  NOT be decorated with the classmethod decorator. Doing so will result
    #  highly undefined behaviour!
    return type(name, bases, namespace)

  def __init__(cls,
               name: str,
               bases: tuple[type, ...],
               namespace: dict,
               **kwargs) -> type:
    #  The __init__ method is an instance method meaning that the first
    #  argument it receives, is the newly created class returned by the
    #  __new__ method.
    #
    #  The __init__ method on type does nothing, but is still called. This
    #  allows a custom metaclass to perform further initialization after
    #  the class object has been created. This is convenient, because the
    #  descriptor protocol invokes __set_name__ on the variables in the
    #  class body, before the __init__ method is invoked.

  def __call__(cls, *args, **kwargs) -> object:
    #  This method is invoked by 'calling' the class object. Typically,
    #  this creates a new instance of the class. A custom metaclass can
    #  customize this behaviour, for example to create a singleton class.
    #
    #  The default implementation is to create an object by calling the
    #  __new__ method on the class object. Please note, that classes are
    #  able to define __new__ and __init__ methods regardless of their
    #  metaclass.
    self = cls.__new__(cls, *args, **kwargs)
    if isinstance(self, cls):
      cls.__init__(self, *args, **kwargs)
    return self

  def __instancecheck__(cls, instance: object) -> bool:
    #  This method is invoked whenever the builtin 'isinstance' function
    #  is called to check if an object is an instance of a class. The default
    #  implementation compares the __class__ attribute of the instance
    #  against the class object and its bases.
    return True if instance.__class__ in cls.mro() else False

  def __subclasscheck__(cls, subclass: type) -> bool:
    #  This method is invoked whenever the builtin 'issubclass' function
    #  is called. It allows classes derived from the metaclass to determine
    #  if another class is a subclass of it. The default implementation
    #  checks if the potential subclass is in the mro of the class object.
    #  Please note, that by default this method returns True if subclass
    #  is the class object itself. This means that in python, classes are
    #  understood to be subclasses of themselves.
    return True if subclass in cls.mro() else False

  def __subclasses__(cls) -> list[type]:
    #  This method returns a list of all immediate subclasses of the given
    #  class 'cls'. Please note that the default implementation does not work
    #  recursively, meaning that subclasses of subclasses are not included.
    #  Further, the default implementation makes use of a list managed
    #  internally by the Python interpreter itself for internal
    #  bookkeeping purposes.
    return type.__subclasses__(cls)

When a class is created, the following events occur:
0.  The metaclass is created.
1.  The __prepare__ method on the metaclass is called creating the
namespace object.
2.  Each non-empty line in the class body is split into key and value pairs
around the equal sign and passed to the namespace object. This object must be
an instance of dict or a subclass of dict. The __setitem__ method on the
namespace object is then invoked on the key and value:
namespace[key] = value or namespace.__setitem__(key, value).
**IMPORTANT**: Under certain circumstances, a line will not contain an
equal sign. In this case, the __getitem__ method on the namespace object
is called:
namespace[key] or namespace.__getitem__(key)
In this case, the namespace object must raise a KeyError on the key
received. This KeyError is then caught and processed internally by the
interpreter. If the namespace object does not raise a KeyError, HIGHLY
UNDEFINED BEHAVIOUR will result.
3.  The resulting namespace object is passed to the __new__ method along
with the name and base classes of the class along with keyword arguments.
4.  The __new__ method creates and returns the class object.
5.  For each instance in the class body, the __set_name__ method is called.
6.  The __init__ method on the metaclass is called with the class object
as the first argument along with the name, base classes and namespace
object. After this method, the class is created.


"""
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from ._type_names import Bases, Namespace, Space
from ._type_names import FunctionDecorator, ClassDecorator, Decorator
from ._base_object import BaseObject
from ._abstract_namespace import AbstractNamespace
from ._abstract_metaclass import AbstractMetaclass
from ._singleton_metaclass import SingletonMetaclass
from ._zeroton_metaclass import ZerotonMetaclass



"""KeeNuMeta provides the metaclass for the KeeNum class."""
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING, Optional, Iterable

from worktoy.keenum import KeeNumSpace, KeeNumObject
from worktoy.meta import AbstractMetaclass
from worktoy.parse import maybe
from worktoy.text import typeMsg, stringList


class KeeNumMeta(AbstractMetaclass):
  """KeeNumMeta provides the metaclass for the KeeNum class."""
  __keenum_entries__ = None
  __allow_instantiation__ = None
  __iter_contents__ = None

  def _getKeeNumEntries(cls, ) -> list[KeeNumObject]:
    """Getter-function for the KeeNum entries."""
    entries = maybe(cls.__keenum_entries__, [])
    if isinstance(entries, list):
      for item in entries:
        if not isinstance(item, KeeNumObject):
          e = typeMsg('item', item, KeeNumObject)
          raise TypeError(e)
      return entries
    e = typeMsg('__keenum_entries__', entries, list)
    raise TypeError(e)

  @classmethod
  def __prepare__(mcls, name: str, _, **kwargs) -> KeeNumSpace:
    """The __prepare__ method is invoked before the class is created."""
    return KeeNumSpace(mcls, name, (), **kwargs)

  def __new__(mcls, name: str, _, keeNumSpace: KeeNumSpace, **__) -> type:
    """The __new__ method is invoked to create the class."""
    namespace = keeNumSpace.compile()
    bases = (KeeNumObject,)
    return AbstractMetaclass.__new__(mcls, name, bases, namespace, )

  def __init__(cls, name: str, _, keeNumSpace: KeeNumSpace, **__) -> None:
    """The __init__ method is invoked after the class is created."""
    cls.__allow_instantiation__ = True
    entries = keeNumSpace.getKeeNumEntries()
    newEntries = []
    for entry in entries:
      newEntries.append(cls(entry))
    cls.__keenum_entries__ = newEntries
    cls.__allow_instantiation__ = False
    AbstractMetaclass.__init__(cls, name, (), keeNumSpace, )

  def __call__(cls, *args, **kwargs) -> object:
    """The __call__ method is invoked when the class is called."""
    if cls.__allow_instantiation__:
      return super().__call__(*args, **kwargs)
    return cls._parse(*args, **kwargs)

  def __getitem__(cls, item: object) -> object:
    """The __getitem__ method is invoked when the class is indexed."""
    args, kwargs = (), {}
    if not isinstance(item, (tuple, list)):
      return cls._parse(item)
    if len(item) == 1:
      return cls.__getitem__(item[0])
    if len(item) > 1:
      if isinstance(item[-1], dict):
        args = (*item[:-1],)
        kwargs = item[-1]
      else:
        args = (*item,)
        kwargs = {}
    return cls._parse(*args, **kwargs)

  def _recognizeIndex(cls, index: int) -> object:
    """The recognize method identifies the entry of the given name. If
    None are found, a ValueError is raised. """
    if index < 0 or index >= len(cls):
      return cls._recognizeIndex(index % len(cls))
    for entry in cls._getKeeNumEntries():
      if int(entry) == index:
        return entry
    e = """Unable to recognize KeeNum entry!"""
    raise ValueError(e)

  def _recognizeName(cls, name: str) -> object:
    """The recognize method identifies the entry of the given name. If
    None are found, a ValueError is raised. """
    for entry in cls._getKeeNumEntries():
      if entry.name.lower() == name.lower():
        return entry
    e = """Unable to recognize KeeNum entry!"""
    raise ValueError(e)

  def _parseKwargs(cls, **kwargs) -> object:
    """Parses keyword arguments"""
    if TYPE_CHECKING:
      assert issubclass(cls, KeeNumObject)
    nameKeys = stringList("""name, key, entry""")
    for key in nameKeys:
      if key in kwargs:
        val = kwargs[key]
        if isinstance(val, cls):
          return val
        if isinstance(val, str):
          return cls._recognizeName(val)
        if isinstance(val, int):
          return cls._recognizeIndex(val)
        if isinstance(val, KeeNumObject):
          return cls._recognizeName(val.name)
        e = typeMsg('KeeNumObject', val, KeeNumObject)
        raise TypeError(e)

  def _parseArgs(cls, *args) -> Optional[object]:
    """Parses positional arguments"""
    for arg in args:
      if isinstance(arg, cls):
        return arg
      if isinstance(arg, KeeNumObject):
        return cls._recognizeName(arg.name)
      if isinstance(arg, str):
        return cls._recognizeName(arg)
      if isinstance(arg, int):
        return cls._recognizeIndex(arg)

  def _parse(cls, *args, **kwargs) -> object:
    """The recognize method identifies the entry indicated by the
    arguments. """
    keeNumObject = maybe(cls._parseArgs(*args), cls._parseKwargs(**kwargs))
    if keeNumObject is None:
      e = """Unable to recognize KeeNum entry!"""
      raise ValueError(e)
    if isinstance(keeNumObject, cls):
      return keeNumObject
    e = typeMsg('keeNumObject', keeNumObject, cls)
    raise TypeError(e)

  def __getattr__(cls, key: str | int) -> object:
    """The __getattr__ method is invoked when an attribute is accessed."""
    if isinstance(key, int):
      return cls._recognizeIndex(key)
    if isinstance(key, str):
      return cls._recognizeName(key)
    return object.__getattribute__(cls, key)

  def __len__(cls) -> int:
    """The length of the class is the number of entries"""
    return len(cls._getKeeNumEntries())

  def __contains__(cls, item: object) -> bool:
    """The __contains__ method is invoked when the class is checked for
    membership."""
    try:
      cls._parse(item)
      return True
    except ValueError:
      return False

  def __iter__(cls) -> object:
    """The __iter__ method is invoked when the class is iterated."""
    cls.__iter_contents__ = [*cls._getKeeNumEntries(), ]
    return cls

  def __next__(cls) -> object:
    """The __next__ method is invoked when the class is iterated."""
    try:
      return cls.__iter_contents__.pop(0)
    except IndexError:
      raise StopIteration


class KeeNum(metaclass=KeeNumMeta):
  """KeeNum provides the metaclass for the KeeNum class."""


if TYPE_CHECKING:
  assert isinstance(KeeNum, Iterable)



"""The 'worktoy.yolo' package provides tools for testing functions and
running tests."""
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from ._yolo import yolo
from ._run_tests import runTests



"""AttriBox subclasses the TypedDescriptor class and incorporates
syntactic sugar for setting the inner class, and for the inner object
creation. """
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

import sys

from typing import TYPE_CHECKING, Callable, Never

from worktoy.desc import TypedDescriptor, Instance, Owner
from worktoy.text import typeMsg, monoSpace

if sys.version_info.minor < 11:
  from typing_extensions import Self
else:
  from typing import Self

if TYPE_CHECKING:
  pass
else:
  AttriClass = object


class AttriBox(TypedDescriptor):
  """AttriBox subclasses the TypedDescriptor class and incorporates
  syntactic sugar for setting the inner class, and for the inner object
  creation. """

  __positional_args__ = None
  __keyword_args__ = None
  __get_callbacks__ = None
  __set_callbacks__ = None
  __del_callbacks__ = None

  def _getGetCallbacks(self) -> list[Callable]:
    """Getter-function for list of functions to be called on get."""
    if self.__get_callbacks__ is None:
      self.__get_callbacks__ = []
    return self.__get_callbacks__

  def _getSetCallbacks(self) -> list[Callable]:
    """Getter-function for list of functions to be called on set."""
    if self.__set_callbacks__ is None:
      self.__set_callbacks__ = []
    return self.__set_callbacks__

  def _getDelCallbacks(self) -> list[Callable]:
    """Getter-function for list of functions to be called on del."""
    if self.__del_callbacks__ is None:
      self.__del_callbacks__ = []
    return self.__del_callbacks__

  def notifyGet(self, callMeMaybe: Callable) -> Callable:
    """Adds given callable to list of callables to be notified on get."""
    self._getGetCallbacks().append(callMeMaybe)
    return callMeMaybe

  def notifySet(self, callMeMaybe: Callable) -> Callable:
    """Adds given callable to list of callables to be notified on set."""
    self._getSetCallbacks().append(callMeMaybe)
    return callMeMaybe

  def notifyDel(self, callMeMaybe: Callable) -> Callable:
    """Adds given callable to list of callables to be notified on del."""
    self._getDelCallbacks().append(callMeMaybe)
    return callMeMaybe

  def ONGET(self, callMeMaybe: Callable) -> Callable:
    """Decorator for adding a function to the get callbacks."""
    return self.notifyGet(callMeMaybe)

  def ONSET(self, callMeMaybe: Callable) -> Callable:
    """Decorator for adding a function to the set callbacks."""
    return self.notifySet(callMeMaybe)

  def ONDEL(self, callMeMaybe: Callable) -> Callable:
    """Decorator for adding a function to the del callbacks."""
    return self.notifyDel(callMeMaybe)

  def __init__(self, *args, **kwargs) -> None:
    """Initializes the AttriBox instance. """
    TypedDescriptor.__init__(self, *args, **kwargs)
    if not kwargs.get('_root', False):
      e = """The AttriBox class should not be instantiated directly!"""
      raise TypeError(e)
    if not args:
      e = """The inner class must be provided. """
      raise TypeError(e)
    innerClass = args[0]
    if not isinstance(innerClass, type):
      e = typeMsg('innerClass', innerClass, type)
      raise TypeError(e)
    self._setInnerClass(innerClass)

  @classmethod
  def __class_getitem__(cls, innerClass: type) -> Self:
    """Syntactic sugar for setting the inner class. """
    return cls(innerClass, _root=True)

  def __call__(self, *args, **kwargs) -> Self:
    """Syntactic sugar for creating an instance of the inner class. """
    self.__positional_args__ = args
    self.__keyword_args__ = kwargs
    return self

  def _getArgs(self, instance: object) -> list:
    """Returns the arguments used to create the inner object. """
    out = []
    for arg in self.__positional_args__:
      if arg is Instance:
        out.append(instance)
      elif arg is Owner:
        out.append(self._getFieldOwner())
      else:
        out.append(arg)
    return out

  def _getKwargs(self, instance: object) -> dict:
    """Returns the keyword arguments used to create the inner object. """
    out = {}
    for (key, value) in self.__keyword_args__:
      if value is Instance:
        out[key] = instance
      elif value is Owner:
        out[key] = self._getFieldOwner()
      else:
        out[key] = value
    return self.__keyword_args__

  def _createInnerObject(self, instance: object) -> object:
    """Creates an instance of the inner class. """
    innerClass = self._getInnerClass()
    args, kwargs = self._getArgs(instance), self._getKwargs(instance)
    innerObject = innerClass(*args, **kwargs)
    if TYPE_CHECKING:
      assert isinstance(innerObject, AttriClass)
    innerObject.setOuterBox(self)
    innerObject.setOwningInstance(instance)
    innerObject.setFieldOwner(self._getFieldOwner())
    innerObject.setFieldName(self._getFieldName())
    return innerObject

  def __str__(self, ) -> str:
    try:
      fieldName = self._getFieldName()
      ownerName = self._getFieldOwner().__name__
    except AttributeError as attributeError:
      if 'has not been assigned to a field' not in str(attributeError):
        raise attributeError
      ownerName = '(TBD)'
      fieldName = '(TBD)'
    innerName = self._getInnerClass().__name__
    return '%s.%s: %s' % (ownerName, fieldName, innerName)

  def __repr__(self, ) -> str:
    try:
      fieldName = self._getFieldName()
    except AttributeError as attributeError:
      if 'has not been assigned to a field' not in str(attributeError):
        raise attributeError
      fieldName = '(TBD)'
    innerName = self._getInnerClass().__name__
    args = [*self.__positional_args__, *self.__keyword_args__]
    args = ', '.join([str(arg) for arg in args])
    return '%s = AttriBox[%s](%s)' % (fieldName, innerName, args)

  @classmethod
  def _getOwnerListName(cls) -> str:
    """Returns the name at which the list of attribute instances of this
    type. Please note that this name is not unique to the owner as they
    are in separate scopes."""
    return '__boxes_%s__' % cls.__qualname__

  def __set_name__(self, owner: type, name: str) -> None:
    """Sets the name of the field. """
    ownerListName = self._getOwnerListName()
    TypedDescriptor.__set_name__(self, owner, name)
    existing = getattr(owner, ownerListName, [])
    if existing:
      return setattr(owner, ownerListName, [*existing, self])
    setattr(owner, ownerListName, [self, ])
    oldInitSub = getattr(owner, '__init_subclass__')

    def newInitSub(cls, *args, **kwargs) -> None:
      """Triggers the extra init"""
      oldInitSub(*args, **kwargs)
      self.applyBoxes(cls)

    setattr(owner, '__init_subclass__', classmethod(newInitSub))

  @classmethod
  def applyBoxes(cls, owner: type) -> None:
    """Applies the boxes to the owner class."""
    ownerListName = cls._getOwnerListName()
    boxes = getattr(owner, ownerListName, [])
    for box in boxes:
      if not isinstance(box, AttriBox):
        e = typeMsg('box', box, AttriBox)
        raise TypeError(e)
      boxName = box._getFieldName()
      setattr(cls, boxName, box)
      cls.__set_name__(box, owner, boxName)

  def __get__(self, instance: object, owner: type) -> object:
    """The __get__ method is called when the descriptor is accessed via the
    owning instance. """
    value = TypedDescriptor.__get__(self, instance, owner)
    for callback in self._getGetCallbacks():
      callback(instance, value)
    return value

  def __set__(self, instance: object, value: object) -> None:
    """The __set__ method is called when the descriptor is assigned a value
    via the owning instance. """
    pvtName = self._getPrivateName()
    oldValue = getattr(instance, pvtName, None)
    setattr(instance, pvtName, value)
    for callback in self._getSetCallbacks():
      callback(instance, oldValue, value)

  def __delete__(self, instance: object, ) -> Never:
    """Deleter-function not yet implemented!"""
    e = """Tried deleting the '%s' attribute from instance of class '%s', 
    but this deleter-function is not yet implemented!"""
    msg = e % (self._getFieldName(), instance.__class__.__name__)
    raise NotImplementedError(monoSpace(msg))



"""Owner is a Zeroton object indicating the owner of the descriptor
instance.  """
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from worktoy.meta import ZerotonMetaclass


class Owner(metaclass=ZerotonMetaclass):
  """Owner is a Zeroton object indicating the class owning the descriptor
  instance.  """


class Instance(metaclass=ZerotonMetaclass):
  """Instance is a Zeroton object indicating the instance owning the
  descriptor instance.  """



"""The 'worktoy.keenum' package provides the KeeNum class for enumeration!"""
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from ._keenum_object import KeeNumObject
from ._keenum_space import KeeNumSpace, auto
from ._keenumeta import KeeNumMeta, KeeNum



"""AbstractDescriptor provides a baseclass for classes implementing the
descriptor protocol."""
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from abc import abstractmethod

from worktoy.text import monoSpace, typeMsg


class AbstractDescriptor:
  """AbstractDescriptor provides a baseclass for classes implementing the
  descriptor protocol."""

  __field_name__ = None
  __field_owner__ = None

  def __init__(self, *__, **_) -> None:
    """Why are we still here?"""
    object.__init__(self)

  def __init_subclass__(cls, **kwargs) -> None:
    """Just to suffer? Or to raise errors?"""
    super().__init_subclass__()

  def _getFieldName(self) -> str:
    """Returns the name of the field the descriptor is assigned to. """
    if self.__field_name__ is None:
      e = """The descriptor has not been assigned to a field. """
      raise AttributeError(monoSpace(e))
    if isinstance(self.__field_name__, str):
      return self.__field_name__
    e = typeMsg('__field_name__', self.__field_name__, str)
    raise TypeError(monoSpace(e))

  def _getFieldOwner(self) -> type:
    """Returns the type of the class the descriptor is assigned to. """
    if self.__field_owner__ is None:
      e = """The descriptor has not been assigned to a field. """
      raise AttributeError(monoSpace(e))
    if isinstance(self.__field_owner__, type):
      return self.__field_owner__
    e = typeMsg('__field_owner__', self.__field_owner__, type)
    raise TypeError(monoSpace(e))

  def _getPrivateName(self, ) -> str:
    """Returns the name of the private attribute used to store the inner
    object. """
    return '__%s_value__' % (self.__field_name__,)

  def __set_name__(self, owner: type, name: str) -> None:
    """The __set_name__ method is called when the descriptor is assigned to
    a class attribute. """
    self.__field_name__ = name
    self.__field_owner__ = owner

  @abstractmethod
  def __instance_get__(self, instance: object, owner: type) -> object:
    """The __instance_get__ method is called when the descriptor is accessed
    via the owning instance. """

  def __get__(self, instance: object, owner: type) -> object:
    """The __get__ method is called when the descriptor is accessed via the
    owning instance. Subclasses should not override this method, but should
    instead implement the __instance_get__ method. """
    if instance is None:
      return self
    return self.__instance_get__(instance, owner)

  def __set__(self, instance: object, value: object) -> None:
    """The __set__ method is called when the descriptor is assigned a value
    via the owning instance. The default implementation raises an error."""
    e = """The '%s' descriptor class does not implement a setter!"""
    raise TypeError(e % self.__class__.__name__)

  def __delete__(self, instance: object) -> None:
    """The __delete__ method is called when the descriptor is deleted via
    the owning instance. The default implementation raises an error."""
    e = """The '%s' descriptor class does not implement a deleter!"""
    raise TypeError(e % self.__class__.__name__)
